from synapse.app.homeserver import setup, run as main_run

CONFIG_PATH = "./tests/example_cfg.yml"
def run() -> None:
    hs = setup(config_options=["-c", CONFIG_PATH, "--generate-missing-and-run"])
    print("Service available at http://localhost:8091/_synapse/client/super_invites")
    print("Shared secret is: exampleLove")
    main_run(hs)

run()


from sqlalchemy import func, select
from synapse.http.servlet import parse_json_object_from_request, parse_string
from synapse.http.site import SynapseRequest
from synapse.types import JsonDict, Tuple  # type: ignore[attr-defined]

from synapse_super_invites.model import Room, Token

from .base import SuperInviteResourceBase, can_edit_token, serialize_token, token_query


class TokensResource(SuperInviteResourceBase):
    async def _async_render_DELETE(
        self, request: SynapseRequest
    ) -> Tuple[int, JsonDict]:
        requester = await self.api.get_user_by_req(request, allow_guest=False)

        token_id = parse_string(request, "token", required=True)
        # query for a specific token
        with self.db.begin() as session:
            token = session.scalar(token_query(token_id))
            if not token:
                return 404, {"error": "Token not found", "errcode": "NOT_FOUND"}
            if not can_edit_token(token, requester):
                return 403, {"error": "Permission denied", "errcode": ""}

            token.deleted_at = func.now()
            session.flush()

        return 200, {}

    async def _async_render_GET(self, request: SynapseRequest) -> Tuple[int, JsonDict]:
        requester = await self.api.get_user_by_req(request, allow_guest=False)

        token_id = parse_string(request, "token")
        # query for a specific token
        if token_id:
            token_data = None
            with self.db.begin() as session:
                token = session.scalar(token_query(token_id))
                if not token:
                    return 404, {"error": "Token not found", "errcode": "NOT_FOUND"}
                if not can_edit_token(token, requester):
                    return 403, {"error": "Permission denied", "errcode": ""}

                token_data = serialize_token(token)

            return 200, {"token": token_data}

        # by default, we list all tokens
        tokens = []
        with self.db.begin() as session:
            for token in session.scalars(
                select(Token).where(
                    Token.owner == str(requester.user),
                    Token.deleted_at == None,  # noqa: E711
                )
            ).all():
                tokens.append(serialize_token(token))
        return 200, {"tokens": tokens}

    async def _async_render_POST(self, request: SynapseRequest) -> Tuple[int, JsonDict]:
        requester = await self.api.get_user_by_req(request, allow_guest=False)
        payload = parse_json_object_from_request(request)
        token_id = payload.get("token", None)
        create_dm = payload.get("create_dm", False)

        token_data = None
        with self.db.begin() as session:
            rooms = [
                session.merge(Room(nameOrAlias=key)) for key in payload.get("rooms", [])
            ]

            token = None
            if token_id:
                token = session.scalar(select(Token).where(Token.token == token_id))

                if token:
                    # check if we have the permission
                    if not can_edit_token(token, requester):
                        return 403, {"error": "Permission denied", "errcode": ""}

                    # token.rooms = rooms
                    token.create_dm = create_dm
                    token.rooms = rooms

            if not token:
                token = Token(
                    token=token_id,
                    create_dm=create_dm,
                    owner=str(requester.user),
                    rooms=rooms,
                )
                session.add(token)

            session.flush()

            token_data = serialize_token(token)

        if self.config.generate_registration_token:
            token_id = token_data["token"]
            # FIXME: it'd be great if we didn't have to resort to using internal args...
            if not (await self.api._store.registration_token_is_valid(token_id)):
                await self.api._store.create_registration_token(
                    token=token_id, uses_allowed=None, expiry_time=None
                )

        return 200, {"token": token_data}



from asyncio import Future
from typing import Any, Awaitable, Dict, Optional, TypeVar
from unittest.mock import Mock

import attr
from synapse.module_api import ModuleApi

from synapse_super_invites import SynapseSuperInvites

TV = TypeVar("TV")


def make_awaitable(result: TV) -> Awaitable[TV]:
    """
    Makes an awaitable, suitable for mocking an `async` function.
    This uses Futures as they can be awaited multiple times so can be returned
    to multiple callers.
    This function has been copied directly from Synapse's tests code.
    """
    future = Future()  # type: ignore
    future.set_result(result)
    return future


@attr.s(auto_attribs=True)
class MockEvent:
    """Mocks an event. Only exposes properties the module uses."""

    sender: str
    type: str
    content: Dict[str, Any]
    room_id: str = "!someroom"
    state_key: Optional[str] = None

    def is_state(self) -> bool:
        """Checks if the event is a state event by checking if it has a state key."""
        return self.state_key is not None

    @property
    def membership(self) -> str:
        """Extracts the membership from the event. Should only be called on an event
        that's a membership event, and will raise a KeyError otherwise.
        """
        membership: str = self.content["membership"]
        return membership


def create_module() -> SynapseSuperInvites:
    # Create a mock based on the ModuleApi spec, but override some mocked functions
    # because some capabilities are needed for running the tests.
    module_api = Mock(spec=ModuleApi)

    # If necessary, give parse_config some configuration to parse.
    config = SynapseSuperInvites.parse_config({})

    return SynapseSuperInvites(config, module_api)



from sqlalchemy import select
from synapse.http.servlet import parse_string
from synapse.http.site import SynapseRequest
from synapse.types import JsonDict, Tuple  # type: ignore[attr-defined]

from synapse_super_invites.model import Accepted

from .base import SuperInviteResourceBase, token_query


class RedeemResource(SuperInviteResourceBase):
    async def _async_render_POST(self, request: SynapseRequest) -> Tuple[int, JsonDict]:
        requester = await self.api.get_user_by_req(request, allow_guest=False)
        my_id = str(requester.user)
        token_id = parse_string(request, "token", required=True)
        invited_rooms = []
        with self.db.begin() as session:
            token = session.scalar(token_query(token_id))
            if not token:
                return 404, {"error": "Token not found", "errcode": "NOT_FOUND"}

            if session.scalar(
                select(Accepted).where(Accepted.user == my_id, Accepted.token == token)
            ):
                return 400, {
                    "error": "Token already redeemed found",
                    "errcode": "ALREADY_REDEEMED",
                }

            if token.owner == my_id:
                return 400, {
                    "error": "Can't redeem your own token",
                    "errcode": "CANT_REDEEM",
                }

            owner = token.owner
            for room in token.rooms:
                await self.api.update_room_membership(
                    sender=owner,
                    target=my_id,
                    room_id=room.nameOrAlias,
                    new_membership="invite",
                )

                await self.api.update_room_membership(
                    sender=my_id,
                    target=my_id,
                    room_id=room.nameOrAlias,
                    new_membership="join",
                )
                invited_rooms.append(room.nameOrAlias)

            if token.create_dm:
                dm_data = await self.api.create_room(
                    my_id, config={"preset": "trusted_private_chat", "invite": [owner]}
                )
                invited_rooms.append(dm_data[0])

            # keep the accepted record
            session.add(Accepted(token=token, user=my_id))
            session.flush()

        return 200, {"rooms": invited_rooms}



from unittest import TestCase

from synapse.config import ConfigError

from synapse_super_invites import SynapseSuperInvites

DEFAULT_CONFIG = {"sql_url": "sqlite:///"}


class ConfigParsingTests(TestCase):
    def test_default_config(self) -> None:
        config = SynapseSuperInvites.parse_config(DEFAULT_CONFIG)
        self.assertEqual(config.generate_registration_token, False)

    def test_fails_no_sql(self) -> None:
        with self.assertRaises(ConfigError):
            SynapseSuperInvites.parse_config({})

    def test_disable_registration_token(self) -> None:
        config = SynapseSuperInvites.parse_config(
            {
                "sql_url": "sqlite:///",
                "generate_registration_token": False,
                "enable_web": False,
            }
        )
        self.assertEqual(config.generate_registration_token, False)



import os

import alembic
import attr
from alembic.config import Config
from sqlalchemy import Engine

PKG_DIR = os.path.dirname(os.path.realpath(__file__))


def run_alembic(engine: Engine) -> None:
    alembic_cfg = Config(os.path.join(PKG_DIR, "alembic.ini"))
    alembic_cfg.set_main_option("script_location", os.path.join(PKG_DIR, "migration"))

    with engine.begin() as connection:
        alembic_cfg.attributes["connection"] = connection
        alembic.command.upgrade(alembic_cfg, "head")


@attr.define
class SynapseSuperInvitesConfig:
    sql_url: str
    generate_registration_token: bool = attr.field(default=False)
    enable_web: bool = attr.field(default=False)



from synapse.http.server import (
    DirectServeHtmlResource,
    finish_request,
    logger,
    set_clickjacking_protection_headers,
)
from synapse.http.site import SynapseRequest
from synapse.module_api import ModuleApi
from synapse.types import Any, JsonDict, Tuple  # type: ignore[attr-defined]


class WebAccessResource(DirectServeHtmlResource):
    def __init__(
        self,
        api: ModuleApi,
    ):
        super().__init__()
        self.api = api

    def _send_response(
        self,
        request: "SynapseRequest",
        code: int,
        response_object: Any,
    ) -> None:
        """Implements _AsyncResource._send_response"""
        # We expect to get bytes for us to write
        assert isinstance(response_object, bytes)
        js_bytes = response_object

        # The response code must always be set, for logging purposes.
        request.setResponseCode(code)

        # could alternatively use request.notifyFinish() and flip a flag when
        # the Deferred fires, but since the flag is RIGHT THERE it seems like
        # a waste.
        if request._disconnected:
            logger.warning(
                "Not sending response to request %s, already disconnected.", request
            )
            return None

        request.setHeader(b"Content-Type", b"text/javascript; charset=utf-8")
        request.setHeader(b"Content-Length", b"%d" % (len(js_bytes),))

        # Ensure this content cannot be embedded.
        set_clickjacking_protection_headers(request)

        request.write(js_bytes)
        finish_request(request)

    async def _async_render_GET(self, request: SynapseRequest) -> Tuple[int, JsonDict]:
        # ensure logged int
        _requester = await self.api.get_user_by_req(request, allow_guest=False)
        access_token = await self.api._auth.get_access_token_from_request(request)
        return 200, 'startApp("{t}")'.format(t=access_token)



from typing import List
from uuid import uuid4

from sqlalchemy import Boolean, Column, DateTime, ForeignKey, String, Table, func
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship


def uuid_short() -> str:
    return str(uuid4()).split("-")[0]


class Base(DeclarativeBase):
    pass


token_rooms = Table(
    "token_room_associations",
    Base.metadata,
    Column("token", ForeignKey("tokens.token"), primary_key=True),
    Column("room", ForeignKey("rooms.nameOrAlias"), primary_key=True),
)


class Token(Base):
    __tablename__ = "tokens"
    token: Mapped[str] = mapped_column(String(50), default=uuid_short, primary_key=True)
    owner: Mapped[str] = mapped_column(String(255))
    create_dm: Mapped[bool] = mapped_column(Boolean)

    rooms: Mapped[List["Room"]] = relationship(
        secondary=token_rooms, back_populates="tokens"
    )

    accepted: Mapped[List["Accepted"]] = relationship(back_populates="token")
    # meta
    created_at = mapped_column(DateTime(timezone=True), server_default=func.now())
    updated_at = mapped_column(
        DateTime(timezone=True), server_default=func.now(), onupdate=func.now()
    )
    deleted_at = mapped_column(DateTime(timezone=True), default=None)

    def __repr__(self) -> str:
        return f"Token({self.token!r}, owned={self.owner!r})"


class Room(Base):
    __tablename__ = "rooms"
    nameOrAlias: Mapped[str] = mapped_column(String(255), primary_key=True)

    tokens: Mapped[List[Token]] = relationship(
        secondary=token_rooms, back_populates="rooms"
    )

    # meta
    created_at = mapped_column(DateTime(timezone=True), server_default=func.now())
    updated_at = mapped_column(
        DateTime(timezone=True), server_default=func.now(), onupdate=func.now()
    )


class Accepted(Base):
    __tablename__ = "accepted"
    id: Mapped[int] = mapped_column(primary_key=True)
    user: Mapped[str] = mapped_column(String(255))

    token_id = mapped_column(ForeignKey("tokens.token"))
    token = relationship("Token", back_populates="accepted")

    # meta
    created_at = mapped_column(DateTime(timezone=True), server_default=func.now())
    updated_at = mapped_column(
        DateTime(timezone=True), server_default=func.now(), onupdate=func.now()
    )



"""Added token table

Revision ID: 25e4ec3cea32
Revises:
Create Date: 2023-11-14 18:52:16.906404

"""
from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "25e4ec3cea32"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "rooms",
        sa.Column("nameOrAlias", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=True,
        ),
        sa.PrimaryKeyConstraint("nameOrAlias"),
    )
    op.create_table(
        "tokens",
        sa.Column("token", sa.String(length=50), nullable=False),
        sa.Column("owner", sa.String(length=255), nullable=False),
        sa.Column("create_dm", sa.Boolean(), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=True,
        ),
        sa.Column("deleted_at", sa.DateTime(timezone=True), nullable=True),
        sa.PrimaryKeyConstraint("token"),
    )
    op.create_table(
        "accepted",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user", sa.String(length=255), nullable=False),
        sa.Column("token_id", sa.String(length=50), nullable=True),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=True,
        ),
        sa.ForeignKeyConstraint(
            ["token_id"],
            ["tokens.token"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "token_room_associations",
        sa.Column("token", sa.String(length=50), nullable=False),
        sa.Column("room", sa.String(length=255), nullable=False),
        sa.ForeignKeyConstraint(
            ["room"],
            ["rooms.nameOrAlias"],
        ),
        sa.ForeignKeyConstraint(
            ["token"],
            ["tokens.token"],
        ),
        sa.PrimaryKeyConstraint("token", "room"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("token_room_associations")
    op.drop_table("accepted")
    op.drop_table("tokens")
    op.drop_table("rooms")
    # ### end Alembic commands ###



from sqlalchemy import func, select
from sqlalchemy.orm import sessionmaker
from synapse.http.server import (
    DirectServeHtmlResource,
    DirectServeJsonResource,
    finish_request,
    logger,
    set_clickjacking_protection_headers,
)
from synapse.http.servlet import parse_json_object_from_request, parse_string
from synapse.http.site import SynapseRequest
from synapse.module_api import ModuleApi
from synapse.types import Any, JsonDict, Requester, Tuple  # type: ignore[attr-defined]

from synapse_super_invites.config import SynapseSuperInvitesConfig
from synapse_super_invites.model import Accepted, Room, Token


def can_edit_token(token: Token, requester: Requester) -> bool:
    # kept outside so we can make it more sophisticated later
    return token.owner == str(requester.user)


def serialize_token(token: Token) -> JsonDict:
    return {
        "token": token.token,
        "create_dm": token.create_dm,
        "accepted_count": len(token.accepted),
        "rooms": [r.nameOrAlias for r in token.rooms],
    }


def token_query(token_id: str):  # type: ignore[no-untyped-def]
    return select(Token).where(
        Token.token == token_id, Token.deleted_at == None  # noqa: E711
    )


class AccessResource(DirectServeHtmlResource):
    def __init__(
        self,
        api: ModuleApi,
    ):
        super().__init__()
        self.api = api

    def _send_response(
        self,
        request: "SynapseRequest",
        code: int,
        response_object: Any,
    ) -> None:
        """Implements _AsyncResource._send_response"""
        # We expect to get bytes for us to write
        assert isinstance(response_object, bytes)
        js_bytes = response_object

        # The response code must always be set, for logging purposes.
        request.setResponseCode(code)

        # could alternatively use request.notifyFinish() and flip a flag when
        # the Deferred fires, but since the flag is RIGHT THERE it seems like
        # a waste.
        if request._disconnected:
            logger.warning(
                "Not sending response to request %s, already disconnected.", request
            )
            return None

        request.setHeader(b"Content-Type", b"text/javascript; charset=utf-8")
        request.setHeader(b"Content-Length", b"%d" % (len(js_bytes),))

        # Ensure this content cannot be embedded.
        set_clickjacking_protection_headers(request)

        request.write(js_bytes)
        finish_request(request)

    async def _async_render_GET(self, request: SynapseRequest) -> Tuple[int, JsonDict]:
        # ensure logged int
        _requester = await self.api.get_user_by_req(request, allow_guest=False)
        access_token = await self.api._auth.get_access_token_from_request(request)
        return 200, 'startApp("{t}")'.format(t=access_token)


class SuperInviteResourceBase(DirectServeJsonResource):
    def __init__(
        self, config: SynapseSuperInvitesConfig, api: ModuleApi, sessions: sessionmaker  # type: ignore[type-arg]
    ):
        super().__init__()
        self.config = config
        self.api = api
        self.db = sessions


class TokensResource(SuperInviteResourceBase):
    async def _async_render_DELETE(
        self, request: SynapseRequest
    ) -> Tuple[int, JsonDict]:
        requester = await self.api.get_user_by_req(request, allow_guest=False)

        token_id = parse_string(request, "token", required=True)
        # query for a specific token
        with self.db.begin() as session:
            token = session.scalar(token_query(token_id))
            if not token:
                return 404, {"error": "Token not found", "errcode": "NOT_FOUND"}
            if not can_edit_token(token, requester):
                return 403, {"error": "Permission denied", "errcode": ""}

            token.deleted_at = func.now()
            session.flush()

        return 200, {}

    async def _async_render_GET(self, request: SynapseRequest) -> Tuple[int, JsonDict]:
        requester = await self.api.get_user_by_req(request, allow_guest=False)

        token_id = parse_string(request, "token")
        # query for a specific token
        if token_id:
            token_data = None
            with self.db.begin() as session:
                token = session.scalar(token_query(token_id))
                if not token:
                    return 404, {"error": "Token not found", "errcode": "NOT_FOUND"}
                if not can_edit_token(token, requester):
                    return 403, {"error": "Permission denied", "errcode": ""}

                token_data = serialize_token(token)

            return 200, {"token": token_data}

        # by default, we list all tokens
        tokens = []
        with self.db.begin() as session:
            for token in session.scalars(
                select(Token).where(
                    Token.owner == str(requester.user),
                    Token.deleted_at == None,  # noqa: E711
                )
            ).all():
                tokens.append(serialize_token(token))
        return 200, {"tokens": tokens}

    async def _async_render_POST(self, request: SynapseRequest) -> Tuple[int, JsonDict]:
        requester = await self.api.get_user_by_req(request, allow_guest=False)
        payload = parse_json_object_from_request(request)
        token_id = payload.get("token", None)
        create_dm = payload.get("create_dm", False)

        token_data = None
        with self.db.begin() as session:
            rooms = [
                session.merge(Room(nameOrAlias=key)) for key in payload.get("rooms", [])
            ]

            token = None
            if token_id:
                token = session.scalar(select(Token).where(Token.token == token_id))

                if token:
                    # check if we have the permission
                    if not can_edit_token(token, requester):
                        return 403, {"error": "Permission denied", "errcode": ""}

                    # token.rooms = rooms
                    token.create_dm = create_dm
                    token.rooms = rooms

            if not token:
                token = Token(
                    token=token_id,
                    create_dm=create_dm,
                    owner=str(requester.user),
                    rooms=rooms,
                )
                session.add(token)

            session.flush()

            token_data = serialize_token(token)

        if self.config.generate_registration_token:
            token_id = token_data["token"]
            # FIXME: it'd be great if we didn't have to resort to using internal args...
            if not (await self.api._store.registration_token_is_valid(token_id)):
                await self.api._store.create_registration_token(
                    token=token_id, uses_allowed=None, expiry_time=None
                )

        return 200, {"token": token_data}


class RedeemResource(SuperInviteResourceBase):
    async def _async_render_POST(self, request: SynapseRequest) -> Tuple[int, JsonDict]:
        requester = await self.api.get_user_by_req(request, allow_guest=False)
        my_id = str(requester.user)
        token_id = parse_string(request, "token", required=True)
        invited_rooms = []
        with self.db.begin() as session:
            token = session.scalar(token_query(token_id))
            if not token:
                return 404, {"error": "Token not found", "errcode": "NOT_FOUND"}

            if session.scalar(
                select(Accepted).where(Accepted.user == my_id, Accepted.token == token)
            ):
                return 400, {
                    "error": "Token already redeemed found",
                    "errcode": "ALREADY_REDEEMED",
                }

            if token.owner == my_id:
                return 400, {
                    "error": "Can't redeem your own token",
                    "errcode": "CANT_REDEEM",
                }

            owner = token.owner
            for room in token.rooms:
                await self.api.update_room_membership(
                    sender=owner,
                    target=my_id,
                    room_id=room.nameOrAlias,
                    new_membership="invite",
                )

                await self.api.update_room_membership(
                    sender=my_id,
                    target=my_id,
                    room_id=room.nameOrAlias,
                    new_membership="join",
                )
                invited_rooms.append(room.nameOrAlias)

            if token.create_dm:
                dm_data = await self.api.create_room(
                    my_id, config={"preset": "trusted_private_chat", "invite": [owner]}
                )
                invited_rooms.append(dm_data[0])

            # keep the accepted record
            session.add(Accepted(token=token, user=my_id))
            session.flush()

        return 200, {"rooms": invited_rooms}



from matrix_synapse_testutils.unittest import (  # type: ignore[import-untyped]
    HomeserverTestCase,
    override_config,
)
from synapse.rest import admin
from synapse.rest.client import login, profile, register, room, sync
from synapse.server import HomeServer
from synapse.types import Dict  # type: ignore[attr-defined]
from synapse.util import Clock
from twisted.test.proto_helpers import MemoryReactor
from twisted.web.resource import Resource

from .test_config import DEFAULT_CONFIG as DEFAULT_MODULE_CFG

DEFAULT_CONFIG = {
    "modules": [
        {
            "module": "synapse_super_invites.SynapseSuperInvites",
            "config": DEFAULT_MODULE_CFG,
        }
    ]
}


# Some more local helpers
class SuperInviteHomeserverTestCase(HomeserverTestCase):  # type: ignore[misc]
    servlets = [
        admin.register_servlets,
        login.register_servlets,
        room.register_servlets,
        profile.register_servlets,
        sync.register_servlets,
        register.register_servlets,
    ]

    def prepare(self, reactor: MemoryReactor, clock: Clock, hs: HomeServer) -> None:
        self.store = hs.get_datastores().main
        self.module_api = hs.get_module_api()
        self.event_creation_handler = hs.get_event_creation_handler()
        self.sync_handler = hs.get_sync_handler()
        self.auth_handler = hs.get_auth_handler()

    def create_resource_dict(self) -> Dict[str, Resource]:
        d: Dict[str, Resource] = super().create_resource_dict()
        for key in self.hs._module_web_resources:
            d[key] = self.hs._module_web_resources[key]
        return d

    # create a room with the given access_token, return the roomId
    def create_room(self, user_id: str) -> str:
        room_id: str = self.get_success(
            self.module_api.create_room(user_id=user_id, config={}, ratelimit=False)
        )[0]
        return room_id


class SimpleInviteTests(SuperInviteHomeserverTestCase):
    @override_config(DEFAULT_CONFIG)  # type: ignore[misc]
    def test_edit_invite_token_rooms(self) -> None:
        m_id = self.register_user("meeko", "password")
        m_access_token = self.login("meeko", "password")

        # this is our new backend.
        channel = self.make_request(
            "GET", "/_synapse/client/super_invites/tokens", access_token=m_access_token
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        self.assertEqual(channel.json_body["tokens"], [])

        # creating five channel
        roomA = self.create_room(m_id)
        roomB = self.create_room(m_id)
        roomC = self.create_room(m_id)
        roomD = self.create_room(m_id)
        roomE = self.create_room(m_id)

        rooms_to_invite = [
            roomB,
            roomC,
            roomD,
        ]
        # create a new one for testing.
        channel = self.make_request(
            "POST",
            "/_synapse/client/super_invites/tokens",
            access_token=m_access_token,
            content={"rooms": rooms_to_invite},
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        token_data = channel.json_body["token"]
        self.assertCountEqual(token_data["rooms"], rooms_to_invite)
        self.assertEqual(token_data["create_dm"], False)
        token = token_data["token"]

        # now we update the roomlist

        rooms_to_invite = [
            roomA,
            roomC,
            roomE,
        ]
        # create a new one for testing.
        channel = self.make_request(
            "POST",
            "/_synapse/client/super_invites/tokens",
            access_token=m_access_token,
            content={"token": token, "create_dm": True, "rooms": rooms_to_invite},
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        token_data = channel.json_body["token"]
        self.assertEqual(token_data["token"], token)
        self.assertCountEqual(token_data["rooms"], rooms_to_invite)
        self.assertEqual(token_data["create_dm"], True)

    @override_config(DEFAULT_CONFIG)  # type: ignore[misc]
    def test_simple_invite_token_test(self) -> None:
        m_id = self.register_user("meeko", "password")
        m_access_token = self.login("meeko", "password")

        # this is our new backend.
        channel = self.make_request(
            "GET", "/_synapse/client/super_invites/tokens", access_token=m_access_token
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        self.assertEqual(channel.json_body["tokens"], [])

        # creating five channel
        _roomA = self.create_room(m_id)
        roomB = self.create_room(m_id)
        roomC = self.create_room(m_id)
        roomD = self.create_room(m_id)
        _roomE = self.create_room(m_id)

        rooms_to_invite = [
            roomB,
            roomC,
            roomD,
        ]
        # create a new one for testing.
        channel = self.make_request(
            "POST",
            "/_synapse/client/super_invites/tokens",
            access_token=m_access_token,
            content={"rooms": rooms_to_invite},
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        token_data = channel.json_body["token"]
        self.assertCountEqual(token_data["rooms"], rooms_to_invite)
        self.assertEquals(token_data["accepted_count"], 0)
        self.assertFalse(token_data["create_dm"])
        token = token_data["token"]

        # redeem the new token

        _f_id = self.register_user("flit", "flit")
        f_access_token = self.login("flit", "flit")

        channel = self.make_request(
            "POST",
            "/_synapse/client/super_invites/redeem?token={token}".format(token=token),
            access_token=f_access_token,
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        # list the rooms we were invited to
        self.assertCountEqual(channel.json_body["rooms"], rooms_to_invite)

        # we see it has been redeemed
        channel = self.make_request(
            "GET",
            "/_synapse/client/super_invites/tokens?token={token}".format(token=token),
            access_token=m_access_token,
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        token_data = channel.json_body["token"]
        self.assertEquals(token_data["accepted_count"], 1)

        # and flit was invited to these, too:
        channel = self.make_request(
            "GET", "/_matrix/client/v3/sync", access_token=f_access_token
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        self.assertEqual(channel.json_body["rooms"].get("invite"), None)
        self.assertCountEqual(
            channel.json_body["rooms"]["join"].keys(), rooms_to_invite
        )

    @override_config(
        {
            "enable_registration": True,
            "registration_requires_token": True,
            "modules": [
                {
                    "module": "synapse_super_invites.SynapseSuperInvites",
                    "config": {
                        "sql_url": "sqlite:///",
                        "generate_registration_token": True,
                    },
                }
            ],
        }
    )  # type: ignore[misc]
    def test_simple_invite_as_registration_token_test(self) -> None:
        _m_id = self.register_user("meeko", "password")
        m_access_token = self.login("meeko", "password")

        # this is our new backend.
        channel = self.make_request(
            "GET", "/_synapse/client/super_invites/tokens", access_token=m_access_token
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        self.assertEqual(channel.json_body["tokens"], [])

        # create a new token for testing.
        channel = self.make_request(
            "POST",
            "/_synapse/client/super_invites/tokens",
            {"rooms": []},
            access_token=m_access_token,
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        token = channel.json_body["token"]

        # let's see if the token exists and is valid
        channel = self.make_request(
            "GET",
            "/_matrix/client/v1/register/m.login.registration_token/validity?token={token}".format(
                token=token["token"]
            ),
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        self.assertTrue(channel.json_body["valid"])

    @override_config(DEFAULT_CONFIG)  # type: ignore[misc]
    def test_simple_invite_token_only_dm_test(self) -> None:
        _m_id = self.register_user("meeko", "password")
        m_access_token = self.login("meeko", "password")

        # this is our new backend.
        channel = self.make_request(
            "GET", "/_synapse/client/super_invites/tokens", access_token=m_access_token
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        self.assertEqual(channel.json_body["tokens"], [])

        # create a new one for testing.
        channel = self.make_request(
            "POST",
            "/_synapse/client/super_invites/tokens",
            access_token=m_access_token,
            content={"rooms": [], "create_dm": True},
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        token_data = channel.json_body["token"]
        self.assertCountEqual(token_data["rooms"], [])
        self.assertEquals(token_data["accepted_count"], 0)
        self.assertTrue(token_data["create_dm"])
        token = token_data["token"]

        channel = self.make_request(
            "GET", "/_synapse/client/super_invites/tokens", access_token=m_access_token
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        self.assertEqual(len(channel.json_body["tokens"]), 1)

        # redeem the new token

        _f_id = self.register_user("flit", "flit")
        f_access_token = self.login("flit", "flit")

        channel = self.make_request(
            "POST",
            "/_synapse/client/super_invites/redeem?token={token}".format(token=token),
            access_token=f_access_token,
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        # only the DM added to
        self.assertEquals(len(channel.json_body["rooms"]), 1)

        new_dm = channel.json_body["rooms"][0]

        # we see it has been redeemed
        channel = self.make_request(
            "GET",
            "/_synapse/client/super_invites/tokens?token={token}".format(token=token),
            access_token=m_access_token,
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        token_data = channel.json_body["token"]
        self.assertEquals(token_data["accepted_count"], 1)

        # and flit was invited to these, too:
        channel = self.make_request(
            "GET", "/_matrix/client/v3/sync", access_token=f_access_token
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        # we are the author of the DM, so we aren't invited, but just added
        self.assertEqual(channel.json_body["rooms"].get("invite"), None)
        self.assertCountEqual(channel.json_body["rooms"]["join"].keys(), [new_dm])

    @override_config(DEFAULT_CONFIG)  # type: ignore[misc]
    def test_simple_invite_token_with_dm_test(self) -> None:
        m_id = self.register_user("meeko", "password")
        m_access_token = self.login("meeko", "password")

        # this is our new backend.
        channel = self.make_request(
            "GET", "/_synapse/client/super_invites/tokens", access_token=m_access_token
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        self.assertEqual(channel.json_body["tokens"], [])

        roomB = self.create_room(m_id)
        roomC = self.create_room(m_id)
        roomD = self.create_room(m_id)

        rooms_to_invite = [
            roomB,
            roomC,
            roomD,
        ]
        # create a new one for testing.
        channel = self.make_request(
            "POST",
            "/_synapse/client/super_invites/tokens",
            access_token=m_access_token,
            content={"rooms": rooms_to_invite, "create_dm": True},
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        token_data = channel.json_body["token"]
        self.assertCountEqual(token_data["rooms"], rooms_to_invite)
        self.assertEquals(token_data["accepted_count"], 0)
        self.assertTrue(token_data["create_dm"])
        token = token_data["token"]

        channel = self.make_request(
            "GET", "/_synapse/client/super_invites/tokens", access_token=m_access_token
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        self.assertEqual(len(channel.json_body["tokens"]), 1)

        # redeem the new token

        _f_id = self.register_user("flit", "flit")
        f_access_token = self.login("flit", "flit")

        channel = self.make_request(
            "POST",
            "/_synapse/client/super_invites/redeem?token={token}".format(token=token),
            access_token=f_access_token,
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        # list the rooms we were invited to
        dm_rooms = []
        for r in channel.json_body["rooms"]:
            if r in rooms_to_invite:
                continue
            dm_rooms.append(r)
        self.assertEqual(len(dm_rooms), 1)

        # we see it has been redeemed
        channel = self.make_request(
            "GET",
            "/_synapse/client/super_invites/tokens?token={token}".format(token=token),
            access_token=m_access_token,
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        token_data = channel.json_body["token"]
        self.assertEquals(token_data["accepted_count"], 1)

        # and flit was invited to these, too:
        channel = self.make_request(
            "GET", "/_matrix/client/v3/sync", access_token=f_access_token
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        self.assertEqual(channel.json_body["rooms"].get("invite"), None)
        self.assertCountEqual(
            channel.json_body["rooms"]["join"].keys(), rooms_to_invite + dm_rooms
        )

    @override_config(DEFAULT_CONFIG)  # type: ignore[misc]
    def test_only_dm_redeem_once(self) -> None:
        _m_id = self.register_user("meeko", "password")
        m_access_token = self.login("meeko", "password")

        # this is our new backend.
        channel = self.make_request(
            "GET", "/_synapse/client/super_invites/tokens", access_token=m_access_token
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        self.assertEqual(channel.json_body["tokens"], [])

        # create a new one for testing.
        channel = self.make_request(
            "POST",
            "/_synapse/client/super_invites/tokens",
            access_token=m_access_token,
            content={"rooms": [], "create_dm": True},
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        token_data = channel.json_body["token"]
        self.assertEquals(token_data["accepted_count"], 0)
        self.assertTrue(token_data["create_dm"])
        token = token_data["token"]

        channel = self.make_request(
            "GET", "/_synapse/client/super_invites/tokens", access_token=m_access_token
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        self.assertEqual(len(channel.json_body["tokens"]), 1)

        # redeem the new token
        _f_id = self.register_user("flit", "flit")
        f_access_token = self.login("flit", "flit")

        channel = self.make_request(
            "POST",
            "/_synapse/client/super_invites/redeem?token={token}".format(token=token),
            access_token=f_access_token,
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        # list the rooms we were invited to
        self.assertEquals(len(channel.json_body["rooms"]), 1)

        # we see it has been redeemed
        channel = self.make_request(
            "GET",
            "/_synapse/client/super_invites/tokens?token={token}".format(token=token),
            access_token=m_access_token,
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        token_data = channel.json_body["token"]
        self.assertEquals(token_data["accepted_count"], 1)

        # trying to redeem again fails.
        channel = self.make_request(
            "POST",
            "/_synapse/client/super_invites/redeem?token={token}".format(token=token),
            access_token=f_access_token,
        )
        self.assertEqual(channel.code, 400, msg=channel.result)

    @override_config(DEFAULT_CONFIG)  # type: ignore[misc]
    def test_deletion(self) -> None:
        _m_id = self.register_user("meeko", "password")
        m_access_token = self.login("meeko", "password")

        # this is our new backend.
        channel = self.make_request(
            "GET", "/_synapse/client/super_invites/tokens", access_token=m_access_token
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        self.assertEqual(channel.json_body["tokens"], [])

        # create a new one for testing.
        channel = self.make_request(
            "POST",
            "/_synapse/client/super_invites/tokens",
            access_token=m_access_token,
            content={"rooms": [], "create_dm": True},
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        token_data = channel.json_body["token"]
        self.assertEquals(token_data["accepted_count"], 0)
        self.assertTrue(token_data["create_dm"])
        token = token_data["token"]

        channel = self.make_request(
            "GET", "/_synapse/client/super_invites/tokens", access_token=m_access_token
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        self.assertEqual(len(channel.json_body["tokens"]), 1)

        # we can access it
        channel = self.make_request(
            "GET",
            "/_synapse/client/super_invites/tokens?token={token}".format(token=token),
            access_token=m_access_token,
        )
        self.assertEqual(channel.code, 200, msg=channel.result)

        # delete it
        channel = self.make_request(
            "DELETE",
            "/_synapse/client/super_invites/tokens?token={token}".format(token=token),
            access_token=m_access_token,
        )
        self.assertEqual(channel.code, 200, msg=channel.result)

        # we can't access it
        channel = self.make_request(
            "GET",
            "/_synapse/client/super_invites/tokens?token={token}".format(token=token),
            access_token=m_access_token,
        )
        self.assertEqual(channel.code, 404, msg=channel.result)

        # and it doesn't show up in the user listing
        channel = self.make_request(
            "GET", "/_synapse/client/super_invites/tokens", access_token=m_access_token
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        self.assertEqual(channel.json_body["tokens"], [])

        # redeem the new token
        _f_id = self.register_user("flit", "flit")
        f_access_token = self.login("flit", "flit")

        # and it can't be redeemed
        channel = self.make_request(
            "POST",
            "/_synapse/client/super_invites/redeem?token={token}".format(token=token),
            access_token=f_access_token,
        )
        self.assertEqual(channel.code, 404, msg=channel.result)

    @override_config(DEFAULT_CONFIG)  # type: ignore[misc]
    def test_cant_redeem_my_own(self) -> None:
        _m_id = self.register_user("meeko", "password")
        m_access_token = self.login("meeko", "password")

        # this is our new backend.
        channel = self.make_request(
            "GET", "/_synapse/client/super_invites/tokens", access_token=m_access_token
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        self.assertEqual(channel.json_body["tokens"], [])

        # create a new one for testing.
        channel = self.make_request(
            "POST",
            "/_synapse/client/super_invites/tokens",
            access_token=m_access_token,
            content={"rooms": [], "create_dm": True},
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        token_data = channel.json_body["token"]
        self.assertEquals(token_data["accepted_count"], 0)
        self.assertTrue(token_data["create_dm"])
        token = token_data["token"]

        channel = self.make_request(
            "GET", "/_synapse/client/super_invites/tokens", access_token=m_access_token
        )
        self.assertEqual(channel.code, 200, msg=channel.result)
        self.assertEqual(len(channel.json_body["tokens"]), 1)

        # and it can't be redeemed
        channel = self.make_request(
            "POST",
            "/_synapse/client/super_invites/redeem?token={token}".format(token=token),
            access_token=m_access_token,
        )
        self.assertEqual(channel.code, 400, msg=channel.result)



import os
from typing import Any, Dict

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from synapse.config import ConfigError
from synapse.module_api import ModuleApi
from twisted.web.static import File

from .config import SynapseSuperInvitesConfig, run_alembic
from .resource import RedeemResource, TokensResource, WebAccessResource

__version__ = "0.8.0"

PKG_DIR = os.path.dirname(os.path.realpath(__file__))


class SynapseSuperInvites:
    def __init__(self, config: SynapseSuperInvitesConfig, api: ModuleApi):
        # Keep a reference to the config and Module API
        engine = create_engine(config.sql_url)
        run_alembic(engine)
        self._sessions = sessionmaker(engine)
        self._api = api
        self._config = config
        self.setup()

    def setup(self) -> None:
        self._api.register_web_resource(
            "/_synapse/client/super_invites/tokens",
            TokensResource(self._config, self._api, self._sessions),
        )
        self._api.register_web_resource(
            "/_synapse/client/super_invites/redeem",
            RedeemResource(self._config, self._api, self._sessions),
        )
        if self._config.enable_web:
            self._api.register_web_resource(
                "/_synapse/client/super_invites/access.js",
                WebAccessResource(self._api),
            )
            self._api.register_web_resource(
                "/_synapse/client/super_invites",
                File(os.path.join(PKG_DIR, "static/")),
            )

    @staticmethod
    def parse_config(config: Dict[str, Any]) -> SynapseSuperInvitesConfig:
        try:
            return SynapseSuperInvitesConfig(**config)
        except TypeError as e:
            raise ConfigError(str(e))
